/*-----------------------------------------------------------------------**
  cdfquantreg: a SAS macro for implementing cdfquantile regression in SAS 
 *-----------------------------------------------------------------------**
	Created by:
	Yiyun Shou and Michael Smithson
	Research School of Psychology
	Australian National University
	Canberra, Australia
	contact: yiyun.shou@anu.edu.au
*-----------------------------------------------------------------------**
   *R package 'cdfquantreg' provides the implementation in R
*-----------------------------------------------------------------------**
Usage:

Argument:
DATA:	A data set including both dependent variables and the independent variables.
		The dependent variable should be within (0, 1) interval. 

DV: 	Specify the variable name for the dependent variable.

FD,SD:	The parent and child distribution of the cdfquantile family distributions. 
		The families and usage are outline in:
		https://cran.rstudio.com/web/packages/cdfquantreg/cdfquantreg.pdf
		
LMIV: 	Specify the names of independent variables in the location submodel.

DMIV: 	Specify the names of independent variables in the dispersion submodel.	

INIT:	The starting values for the parameters, should the starting values for the intercepts of each submodel.
		E.g., the length of INIT should be 2 (1 for location submodel and 1 for dispersion submodel) for an intercept only model.
*/


%macro cdfquantreg(DATA, DV, FD, SD, LMIV= , DMIV= , INIT= );

%let distname = %sysfunc(cat('cdfquantreg Model for ', &FD, ' ', &SD, ' Distribution'));

/*------------------------------------------------*/
/*Process the data set by sorting the variables*/
	data _NULL_;
		if 0 then set &DATA nobs=n;
			call symputx('nrows',n); /*Get the sample size*/
		stop;
	run;

	data &DATA;
	  SET &DATA;
	      _id = _N_; /*Assign subject ID*/ 
	  RUN;

	PROC SORT DATA=&DATA OUT= datatemp ;
	  BY &DV; /*Sort the data based on the dv*/ 
	RUN ;

	DATA datatemp;
	 	SET datatemp nobs = nobs;
	      _rank = _N_;
		  _q =  (_N_-1)*((0.999999 - 1e-06)/(&nrows-1)) + 1e-06; /*Get empirical quantile*/ 
	 RUN;

	PROC SORT DATA = datatemp OUT = DATA; /*Sort the data back to the original order*/
	  BY _id;
	RUN;

/*------------------------------------------------*/
/*Process IVS and parameters*/
	%let lmvars = b0;  /*intiate location model;*/
	%let dmvars = d0;  /*intiate dispersion model;*/

	%IF %length(&LMIV) = 0 %THEN %let numlmvars = 0;
		%ELSE %let numlmvars = %numargs(&LMIV);  /*get number of predictors in lmvars;*/

	%IF %length(&DMIV) = 0 %THEN %let numdmvars = 0;
		%ELSE %let numdmvars = %numargs(&DMIV);  /*get number of predictors in dmvars;*/

	%IF %length(&INIT) = 0 %THEN /*default starting values if none is provided;*/
	%DO;
		%let INIT = 0.1;  
		%let numvars = &numlmvars + &numdmvars + 2; /*No. of parameters = no of predictors + 2*/
		%do j = 2 %to &numvars;
			%let INIT = %STR(&INIT |0.1);
		%end;
		%let parmsb = %STR(b0 = 0.1);
		%let parmsd = %STR(d0 = 0.1); 
	%END;
		%ELSE %DO;
			%let initb0 = %scan(&INIT., 1, |);
			%let initd0 = %scan(&INIT., 2 + &numlmvars, |);
			%let parmsb = %STR(b0 = &initb0);
			%let parmsd = %STR(d0 = &initd0); 		
		%END;

	/*Process the two submodels*/
		%IF %length(&lmiv) ne 0 %THEN 
			%DO;
				%do i=1 %to &numlmvars;
					%let varb = %scan(&LMIV, &i, %str( ));
					%let initb = %scan(&INIT, &i + 1, |);
					%let lmvars = &lmvars + b&i*&varb;/*beta coefficients for location model;*/
					%let parmsb = %STR(&parmsb b&i=&initb);/*initial starting values;*/
				%end;
			%END;

		%IF %length(&dmiv) ne 0 %THEN 
			%DO;
				%do j=1 %to &numdmvars;
					%let vard = %scan(&dmiv, &j, %str( ));
					%let initd = %scan(&INIT, &j + 2 + &numlmvars, |);
					%let dmvars = &dmvars + d&j*&vard;
					%let parmsd = %STR(&parmsd d&j = &initd);/*initial starting values;*/
				%end;
			%END;

	%let parmsall = %STR(&parmsb &parmsd); 

	%put &lmvars;
	%put &dmvars;
	%put &parmsall;
/*-----------------------------------------------------**
	Distribtuion likelihood and quantile functions 
 *-----------------------------------------------------***/
	%Let t2case = ;

/**%INCLUDE tempdir(SAS_CDFS);
*%cdffunc(&FD, &SD); This function returns a log likelihood function and a quantile function*/

/*-----------------------
	Arcsinh - XX models
----------------------*/
	%IF &FD = 'arcsinh' %THEN 
	%DO;
		%IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk =  -log(sigma) + arsinh((-mu - log(1 - y) + log(y))/sigma) - 2 * log(1 + exp(arsinh((-mu - log(1 - y) + log(y))/sigma))) - (1/2) * log(1 + (mu + log(1 - y) - log(y))**2/sigma**2) - log(1 - y) - log(y);
				%Let qqfun =  exp(mu)/(exp(mu) + exp((sigma - 2*_q*sigma)/(2*_q - 2*_q**2)));
			%END;

		%ELSE %IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk = arsinh(((1 - 2*y)/(2*(-1 + y)*y) - mu)/sigma) - 2*log(1 + exp(arsinh(((1 - 2*y)/(2*(-1 + y)*y) - mu)/sigma))) - 
						log(1 - y) - log(y) + log(1 - 2*y + 2*y**2) - (1/2)*log(1 + 4*y*(-1 + mu) + 4*y**4*(mu**2 + sigma**2) + 4*y**2*(1 - 3*mu + mu**2 + sigma**2) - 8*y**3*(-mu + mu**2 + sigma**2));
				%Let qqfun =   1/(1 + exp(-arsinh(((1 - 2*_q)/(2*(-1 + _q)*_q) + mu/sigma)/(1/sigma))));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk =   -log(2) - log(sigma) + arsinh((mu + artanh(1 - 2 * y))/sigma) - 
	      							2 * log(1 + exp(arsinh((mu + artanh(1 - 2 * y))/sigma))) - 
	      							(1/2) * log(1 + (mu + artanh(1 - 2 * y))**2/sigma**2) - log(1 - y) - log(y);
				%Let qqfun =   1/2* (1 + tanh(mu + (sigma - 2 * _q *sigma)/( 2 *(-1 + _q)* _q)));
			%END;

		%ELSE %IF &SD = 'burr8' %THEN
			%DO;
				%Let loglk =  log(constant('pi')) - log(sigma) + arsinh((-mu + log(tan((constant('pi') * y)/2)))/sigma) - 2 * log(1 + 
	       					 exp(arsinh((-mu + log(tan((constant('pi') * y)/2)))/sigma))) + log(1/sin(CONSTANT('pi') * y)) - (1/2) * log(1 + 
	        			(mu - log(tan((constant('pi') * y)/2)))**2/sigma**2);
				%Let qqfun =    2 * (artanh(exp(mu + (sigma - 2 * _q * sigma)/(2 * (-1 + _q) * _q)))/constant('Pi'));
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk =  log(constant('pi')) + ARSINH((mu + cot(constant('pi') * y))/sigma) - 2 * log(1 + exp(ARSINH((mu + cot(constant('pi') * 
	       						 y))/sigma))) - (1/2) * log(mu**2 + sigma**2 + 2 * mu * cot(constant('pi') * y) + cot(constant('pi') * y)**2) + 
	       							 2 * log(1/sin(CONSTANT('pi') * y));
				%Let qqfun =  1/2 + artanh(mu + (sigma - 2 * _q * sigma)/(2 * (-1 + _q) * _q))/constant('Pi');
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let t2case = %NRSTR(if %(%(y >= 0 %) & %(y < 1/2 %)%) then
						a1 = -sqrt%(%(1 - 2 * y %)**2/ %(2 * %( 1 - y %) * y %) %);
							else if %(y >= 1/2 & y <= 1%) then
		 				a1 = sqrt %(%(1 - 2 * y %)**2/%(2 * %(1 - y%) * y%)%);
					else a1 = 0;
						if %(y =0.5%) then
					a2 = 2 * sqrt %( 2 %);
					else if %(y > 0 & y < 1/2 %)then
					 	a2 = %(1 - 2 * y %)/sqrt %(8 * %(1 - 2 * y%)**2 * %(%(1 - y %) * y %)**3 %);
					else if %(%(y > 1/2 & y < 1%)%)then
					 	a2 = %(2 * y - 1%)/sqrt%(8 * %(1 - 2 * y%)**2 * %(%(1 - y%) * y%)**3%);
					else if %(y = 1 | y = 0%)then
					 	a2 = .;
					else a2 =0;);
				%Let loglk =  log((exp(arsinh((-mu + a1)/sigma)) * a2)/((1 + exp(arsinh((-mu + a1)/sigma)))**2 * sigma *  sqrt(1 + (mu - a1)**2/sigma**2)));
				%Let qqfun =  1/2 + (((1 - 2 * _q)/(2 * (-1 + _q) * _q) +  mu/sigma) * sigma)/(2 * sqrt(2 + (((1 - 2 * _q)/(2 * (-1 + _q) * _q) + 
	                                                                                    mu/sigma) * sigma)**2));
			%END;
	%END;

/*-----------------------
	Burr VII - XX models
----------------------*/
	%IF &FD = 'burr7' %THEN 
	%DO;
		%IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk =  -log(4) - log(sigma) + 2 * log(COSH((-mu - (1 - 2 * y)/(2 * (1 - y) * y))/sigma)**(-1)) - 
		        2 * log(1 - y) - 2 * log(y) + log(1 - 2 * (1 - y) * y);
				%Let qqfun =  (-1 + mu - sigma *artanh(1 - 2* _q) + sqrt(1 + mu**2 -2* mu *sigma* artanh(1 - 2*_q) + sigma**2 *artanh( 1 - 2* _q)**2))/( 2 *mu - 2 *sigma* artanh(1 - 2*_q));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk =  2 * log(COSH((mu + artanh(1 - 2 * y))/sigma)**(-1)) - log(4 * sigma * y - 4 * sigma * y**2);
				%Let qqfun =  (1/2 )*(1 + tanh(mu - sigma *artanh(1 - 2*_q)));
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk = -log((2 * sigma)/CONSTANT('pi')) + 2 * log(1/sin(CONSTANT('pi') * y)) + 2 * log(COSH((mu + 1/tan(CONSTANT('pi') * y))/sigma)**(-1));
				%Let qqfun =   1/2 + atan(mu - sigma *artanh(1 - 2 *_q))/CONSTANT('pi');
			%END;

		%ELSE %IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk =  -log(2) - log(sigma) + 2 * log(COSH((-mu - log(1 - y) + log(y))/sigma)**(-1)) - log(1 - y) - log(y);
				%Let qqfun =  exp(mu)/(exp(mu) + exp(sigma* artanh(1 - 2*_q)));
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let loglk =  log((1/COSH(((-1 + 2 * y)/(sqrt(2) * sqrt((1 - y) * y)) - mu)/sigma))**2/(4 * sqrt(2) * (((1 - y) * y)**(3/2)) * sigma));
				%Let qqfun =  1/2 + ((-artanh(1 - 2* _q) + mu/sigma)*sigma)/( 2* sqrt(2 + ((-artanh(1 - 2*_q) + mu/sigma)*sigma)**2));
			%END;
	%END;

/*-----------------------
	Burr VIII - XX models
----------------------*/
	%IF &FD = 'burr8' %THEN 
	%DO;
		%IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk =  (-(1/(2 * sigma))) * (2 * sigma * log(constant('pi')) + 2 * sigma * log(sigma) + 2 * sigma * 
		        log(exp((2 * (mu - 1/(1 - y)))/sigma) + exp(1/(sigma * (-1 + y) * y))) + 4 * sigma * 
		        log(1 - y) + 4 * sigma * log(y) - 2 * sigma * log(1 - 2 * (1 - y) * y) + 2 * (1/(1 - 
		        y)) - 2 * mu * (1/(1 - y)) + 1/((1 - y) * y) + 2 * mu * (y/(1 - y)));
				%Let qqfun =   1/(1 - mu + sigma* log(cot((CONSTANT('pi')*_q)/2)) + sqrt( 1 + mu**2 - 2*mu * sigma*(log(cot((CONSTANT('pi')*_q)/2))) + sigma**2 * (log( cot((CONSTANT('pi')*_q)/2)))**2));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk = -log(2 * constant('pi') * y * sigma - 2 * constant('pi') * y**2 * sigma) + log((COSH((mu + artanh(1 - 2 * y))/sigma))**(-1));
				%Let qqfun =  (1/2) * (1 + tanh(mu - sigma * (log(cot(CONSTANT('pi') *_q/2)))));
			%END;

		%ELSE %IF &SD = 'burr8' %THEN
			%DO;
				%Let loglk =  (mu + sigma * log(2) + sigma * log(1/sin(CONSTANT('pi') * y)) + log(tan(constant('pi') * y/2)) - sigma * log(exp((2 * mu)/sigma) * sigma + sigma * tan((constant('pi') * y)/2)**(2/sigma)))/sigma;
				%Let qqfun =  2 * atan(exp(mu - sigma * log(cot(CONSTANT('pi') * _q/2))))/CONSTANT('pi');
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk = (mu + sigma * log(2) - sigma * log(sigma) + cot(constant('pi') * y) - sigma * log(1 + exp((2 * (mu + cot(constant('pi') * y)))/sigma)) + 2 * sigma * log(1/sin(CONSTANT('pi') * y)))/sigma;
				%Let qqfun =  1/2 + atan(mu - sigma * (log(cot((CONSTANT('pi') * _q)/2))))/CONSTANT('pi');
			%END;

		%ELSE %IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk =  -((-mu - sigma * log(2) + sigma * log(CONSTANT('pi')) + sigma * log(sigma) + (1 + sigma) * 
		        log(1 - y) + (-1 + sigma) * log(y) + sigma * log(exp((2 * mu)/sigma) + y**(2/sigma)/(1 - y)** (2/sigma)))/sigma));
				%Let qqfun =  exp(mu)/(exp(mu)+ cot((CONSTANT('pi')* _q)/2)**sigma);
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let t2case = %NRSTR(if %(y >= 0 & y < 1%) then
			 						a1 = COSH%(%(%(1 - 2 * y%)/%(sqrt%(2%) * sqrt%(%(1 - y%) * y%)%) + mu%)/
									sigma%)**%(-1%)/%(2 * sqrt%(2%) * CONSTANT%('pi'%) * %(%(1 - y%) * y%)**%(3/2%) * sigma%);
									else if %(y = 1 | y = 0%)then
			 							a1 = .;
									else a1 =0;);
				%Let loglk =  log(a1);
				%Let qqfun =  2 * atan(exp(mu - sigma * log(cot(constant('pi') * _q/2))))/constant('pi');
			%END;
	%END;


/*-----------------------
	Cauchy - XX models
----------------------*/
	%IF &FD = 'cauchit' %THEN 
	%DO;
		%IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk =  log((2 * sigma)/constant('pi')) + log(1 + 2 * (-1 + y) * y) - log(4 * sigma**2 * (-1 + y)**2 *y**2 + (-1 + 2 * y + 2 * mu * (-1 + y) * y)**2);
				%Let qqfun =    (-1 + mu - sigma* cot((CONSTANT('pi'))* _q) + sqrt( 1 + mu**2 - 2* mu* sigma* cot((CONSTANT('pi'))* _q) + sigma**2* cot((CONSTANT('pi'))* _q)**2))/(2* mu - 2* sigma* cot((CONSTANT('pi'))* _q));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk = -log(2) - log(constant('pi')) + log(sigma) - log(mu**2 + sigma**2 + 2 * mu * artanh(1 - 2 * 
		        y) + artanh(1 - 2 * y)**2) - log(1 - y) - log(y);
				%Let qqfun =  (1/2) *(1 + tanh(mu - sigma* cot((CONSTANT('pi'))* _q)));
			%END;

		%ELSE %IF &SD = 'burr8' %THEN
			%DO;
				%Let loglk =  log(sigma) + log(1/sin(CONSTANT('pi') * y)) - log(mu**2 + sigma**2 - 2 * mu * log(tan((constant('pi') * y)/2)) + log(tan((constant('pi') * y)/2))**2);
				%Let qqfun =  (2* (atan(exp(mu - sigma* cot((CONSTANT('pi'))* _q)))))/(CONSTANT('pi'));
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk =  log(sigma) - log(mu**2 + sigma**2 + 2*mu/tan(constant('pi')*y) + 1/tan(constant('pi')*y)**2) + 2*log(1/sin(constant('pi')*y));
				%Let qqfun =   1/2 + atan((-(1/tan(CONSTANT('pi')*_q)) + mu/sigma)*sigma)/CONSTANT('pi');
			%END;

		%ELSE %IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk = -log(constant('pi')) + log(sigma) - log(mu**2 + sigma**2 - 2*mu*(-log(1 - y) + log(y)) + (-log(1 - y) + log(y))**2) - log(1 - y) - log(y);
				%Let qqfun =  1/(1 + exp(-mu + sigma*tan(0.5*CONSTANT('pi') - CONSTANT('pi')*_q)));
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let t2case = %NRSTR(
					if %(%(y >= 0%) & %(y < 1/2%)%) then
						a1 = -sqrt%(%(1 - 2 * y%)**2/%(2 * %(1 - y%) * y%)%);
					else if %(y >= 1/2 & y <= 1%) then
		 				a1 = sqrt%(%(1 - 2 * y%)**2/%(2 * %(1 - y%) * y%)%);
					else a1 = 0;
					if %(y =0.5%) then
						a2 = 2 * sqrt%(2%);
					else if %(y > 0 & y < 1/2%)then
		 				a2 = %(1 - 2 * y%)/sqrt%(8 * %(1 - 2 * y%)**2 * %(%(1 - y%) * y%)**3%);
					else if %(%(y > 1/2 & y < 1%)%)then
		 				a2 = %(2 * y - 1%)/sqrt%(8 * %(1 - 2 * y%)**2 * %(%(1 - y%) * y%)**3%);
					else if %(y = 1 | y = 0%)then
		 				a2 = .;
					else a2 =0;
					a3 = a2/%(CONSTANT%('pi'%) * sigma * %(1 + %(mu - a1%)**2/sigma**2%)%););
				%Let loglk =  log(a3);
				%Let qqfun = 1/2 + ((mu/sigma)-cot(CONSTANT('pi')*_q))*sigma/(2*sqrt(2 + (((mu/sigma)-cot(CONSTANT('pi')*_q))*sigma)**2))  ;
			%END;
	%END;


/*-----------------------
	Logit - XX models
----------------------*/
	%IF &FD = 'logit' %THEN 
	%DO;
		%IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk =  ((mu-sigma*log(sigma))-(-1+sigma)*log(-1+1/y)-2*sigma*log(y+exp(mu/sigma)*(-1+1/y)**(1/sigma)*y))/sigma;
				%Let qqfun =  1/(1 + exp((-(-log(-1 + 1/_q) + mu/sigma)) *sigma));
			%END;

		%ELSE %IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk =  -log(2) - log(sigma) - 2 * log(exp(1/(sigma - sigma * y)) + exp(mu/sigma + 
						1/(2 * sigma * y - 2 * sigma * (y**2)))) - 2 * log(1 - y) - 2 * log(y) + log(1 - 2 * y + 2 * 
	        			(y**2)) + 1/(2 * sigma * y - 2 * sigma * (y**2)) + 2 * (y/(2 * sigma * y - 2 * sigma * (y**2))) + 
	        			2 * mu * (y/(2 * sigma * y - 2 * sigma * (y**2))) - 2 * mu * ((y**2)/(2 * sigma * y - 2 * sigma * (y**2)));
				%Let qqfun =  1/(1 + exp(-arsinh((-log(-1 + 1/_q) + mu/sigma)*sigma)));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk =  2 * log(COSH((mu + artanh(1 - 2 * y))/(2 * sigma))**(-1)) - log(8 * sigma * y - 8 * sigma * y**2);
				%Let qqfun =  (1/2 )*(1 + tanh(mu - sigma* (log(-1 + 1/_q))));
			%END;

		%ELSE %IF &SD = 'burr8' %THEN
			%DO;
				%Let loglk =  log(1/sin(CONSTANT('pi') * y)) + ((mu + sigma * log(constant('pi')) - sigma * log(sigma)) + 
								log(tan((constant('pi') * y)/2)) - 2 * sigma * log(exp(mu/sigma) + tan((constant('pi') * y)/2)**(1/sigma)))/sigma;
				%Let qqfun =  (2 * (atan(exp(mu - sigma * (log(-1 + 1/_q))))))/constant('Pi');
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk =  (mu + sigma * log(constant('pi')) - sigma * log(sigma) + cot(constant('pi') * y) - 2 * sigma * log(1 + exp((mu + cot(constant('pi') * y))/sigma)) + 2 * sigma * log(1/sin(CONSTANT('pi') * y)))/sigma;
				%Let qqfun = 1/2 + atan((-log(-1 + 1/_q) + mu/sigma) *  sigma)/constant('Pi');
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let t2case = %NRSTR(if %(%(y >= 0 %) & %(y < 1/2 %)%) then
							a1 = -sqrt%(%(1 - 2 * y %)**2/ %(2 * %( 1 - y %) * y %) %);
								else if %(y >= 1/2 & y <= 1%) then
			 				a1 = sqrt %(%(1 - 2 * y %)**2/%(2 * %(1 - y%) * y%)%);
						else a1 = 0;
							if %(y =0.5%) then
						a2 = 2 * sqrt %( 2 %);
						else if %(y > 0 & y < 1/2 %)then
						 	a2 = %(1 - 2 * y %)/sqrt %(8 * %(1 - 2 * y%)**2 * %(%(1 - y %) * y %)**3 %);
						else if %(%(y > 1/2 & y < 1%)%)then
						 	a2 = %(2 * y - 1%)/sqrt%(8 * %(1 - 2 * y%)**2 * %(%(1 - y%) * y%)**3%);
						else if %(y = 1 | y = 0%)then
						 	a2 = .;
						else a2 =0;);
				%Let loglk =  log((exp((mu + a1)/sigma) * a2)/((exp(mu/sigma) + exp(a1/sigma))**2 * sigma));
				%Let qqfun =  (1/2 )*(1 + tanh(mu - sigma* (log(-1 + 1/_q))));
			%END;
	%END;

/*------------------------------
	T2 - XX models
----------------------*/
	%IF &FD = 't2' %THEN 
	%DO;
	%Let t2case = %NRSTR(if %(%(_q >= 0%) & %(_q < 1/2%)%) then
						a1 = -sqrt%(%(1 - 2 * _q%)**2/%(2 * %(1 - _q%) * _q%)%);
						else if %(_q >= 1/2 & _q <= 1%) then
			 			a1 = sqrt%(%(1 - 2 * _q%)**2/%(2 * %(1 - _q%) * _q%)%););
		%IF &SD = 'arcsinh' %THEN
			%DO;
				%Let loglk = log(4) + 2 * log(sigma) + 3 * log(1 - y) + log(y) - (3/2) * log(1 + 4 * (-1 + 
		        mu) * y + 4 * (1 - 3 * mu + mu**2 + 2 * sigma**2) * y**2 - 8 * (-mu + mu**2 + 2 * sigma**2) * 
		        y**3 + 4 * (mu**2 + 2 * sigma**2) * y**4);
				%Let qqfun =  1/(1 + exp(-arsinh(mu + sigma*a1))));
			%END;

		%ELSE %IF &SD = 'burr7' %THEN
			%DO;
				%Let loglk = -log(2) + 2 * log(sigma) - (3/2) * log(mu**2 + 2 * sigma**2 + 2 * mu * artanh(1 - 
		        2 * y) + artanh(1 - 2 * y)**2) - log(1 - y) - log(y);
				%Let qqfun = (tanh((a1 + mu/sigma) * sigma) + 1)/2;
			%END;

		%ELSE %IF &SD = 'burr8' %THEN
			%DO;
				%Let loglk = (log(constant('pi')) + 2 * log(sigma)) + log(1/sin(CONSTANT('pi') * y)) - (3/2) * log(mu**2 + 2 * sigma**2 - 
		        2 * mu * log(tan((constant('pi') * y)/2)) + log(tan((constant('pi') * y)/2))**2);
				%Let qqfun =  2 * (atan(exp((a1 + mu/sigma) * sigma))/constant('Pi'));
			%END;

		%ELSE %IF &SD = 'cauchy' %THEN
			%DO;
				%Let loglk = log(constant('pi')) + log(sigma) - log(mu**2 + 2 * sigma**2 + 2 * mu * cot(constant('pi') * y) + cot(constant('pi') * 
		        y)**2) - (1/2) * log(2 + (mu + cot(constant('pi') * y))**2/sigma**2) + 2 * log(1/sin(CONSTANT('pi') * y));
				%Let qqfun =  1/2 + atan((a1 + mu/sigma) * sigma)/constant('pi');
			%END;

		%ELSE %IF &SD = 'logistic' %THEN
			%DO;
				%Let loglk = log(sigma) -1/2 * log(2 + (mu + log(1 - y) - log(y))**2/sigma**2) -log(mu**2 + 
		  		 2* sigma**2 -2* mu *(-log(1 - y) + log(y)) + (-log(1 - y) + log(y))**2) -log(1 - y) - log(y);
				%Let qqfun = 1/(1 + exp(-mu - sigma*a1));
			%END;

		%ELSE %IF &SD = 't2' %THEN
			%DO;
				%Let loglk = 2*log(sigma) - (3/2)*log(1 + 2*sqrt(2)*mu*sqrt(1 - y)*sqrt(y) + 2*(-2 + mu**2 + 2*sigma**2)*y - 4*mu*sqrt(2 - 2*y)*y**(3/2) - 2*(-2 + mu**2 + 2*sigma**2)*y**2);
				%Let qqfun = 1/2 + (a1 * sigma + mu)/(2 * sqrt(2 + (a1 * sigma + mu)**2));
			%END;
	%END;


/*Fit the model*/
proc nlmixed data = data tech = trureg hess cov itdetails;
	parms &parmsall;
	title &distname;
	y = &DV;
	xb = &lmvars;
	mu = xb;
	wd = &dmvars; 
	sigma = exp(wd);
	&t2case;
	ll = &loglk; 
	model y ~ general(ll);
	predict mu out = predmu;
	predict sigma out = predsigma;
	predict &qqfun out = predy; 
	ODS OUTPUT ParameterEstimates =_pe0;
run;

/*Produce output including predicting values the model*/
data dataout; 
	merge &data predmu(keep = Pred rename= (Pred = predmu))
		  predsigma(keep = Pred rename= (Pred=predsigma))
		  predy(keep = Pred rename= (Pred=predy));
	LABEL predmu 	= "Predicted Mean"
		  predsigma = "Predicted Sigma"
		  predy 	= "Predicted Value";
run;

/*Process residuals*/
Proc sql;
	Create table dataout as
	Select *, std(predy) as std_var
From dataout;

data dataout;
		set dataout ;
		resid= predy - Y;
		pearsonresid = (predy - Y)/std_var;
run;

data dataout;
	set dataout(drop=std_var);
run;

*Delete temporary output;
proc datasets nolist noprint;
	delete datatemp predmu predsigma predy;
run; 

%mend cdfquantreg;

/*help function
From: Voronca, D. C., Egede, L. E., & Gebregziabher, M. ANALYSIS OF ZERO INFLATED LONGITUDINAL DATA USING PROC NLMIXED.*/
%macro numargs(arg, delimit);
 %if %quote(&arg)= %then %do;
 0
 %end;
 %else %do;
 %let n=1;
 %do %until (%qscan(%quote(&arg), %eval(&n), %str( ))=%str());
 %let n=%eval(&n+1);
 %end;
%eval(&n-1)
 %end;
 %mend numargs;
